"""Regressions for the acdd-v2 audit fixes."""

from __future__ import annotations

import io
import json
import sys
from pathlib import Path

import pytest

from acdd._doc import command_outcome_ok, sha256
from acdd.adapter import AdapterError, load_adapter
from acdd.cli import main
from acdd.fingerprint import fingerprint_for_gate, fingerprint_gate
from acdd.model import (
    AcddError,
    Check,
    Document,
    Gate,
    Profile,
    Subtask,
    load_document,
    load_profile,
)
from acdd.validate import validate


def _args(doc: Path, profile: Path, adapter: Path) -> list[str]:
    return [
        str(doc),
        str(profile),
        "--workspace-root",
        str(doc.parent),
        "--adapter",
        f"implementation={adapter}",
    ]


def _review_raw(
    *, reviewer_uuid: str = "00000000-0000-4000-8000-000000000002", raw: str = '{"findings": []}'
) -> dict:
    return {"type": "review_raw", "reviewerSessionUuid": reviewer_uuid, "raw": raw}


def test_basis_artifact_with_failed_outcome_is_rejected(tmp_path: Path):
    (tmp_path / "a.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: x
role: implementation
gates:
  build/v1:
    checks:
      c: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )
    adapter = load_adapter(tmp_path / "a.yaml")
    artifact = tmp_path / "basis.jsonl"
    artifact.write_text(
        json.dumps(
            {
                "type": "command_run",
                "evidenceId": "e1",
                "gate": "build/v1",
                "check": "c",
                "exitCode": 1,
            }
        )
        + "\n",
        encoding="utf-8",
    )
    gate = Gate(
        "build/v1", "implementation", (Check("c", "basis", "success"),), ("source",), ("pass",)
    )
    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "task.md"
    )
    fingerprint = fingerprint_for_gate(document, gate, tmp_path, adapter)
    child = {
        "kind": "basis",
        "id": "e1",
        "gate": "build/v1",
        "check": "c",
        "issuerRole": "implementation",
        "artifactSha256": sha256(artifact),
        "inputFingerprint": fingerprint,
        "basisRef": "basis.jsonl",
        "scope": [],
        "classifiedRefs": [],
    }
    bundle = {
        "kind": "bundle",
        "id": "b",
        "gate": "build/v1",
        "issuerRole": "implementation",
        "checkEvidence": ["e1"],
        "inputFingerprint": fingerprint,
    }
    document.evidence = [child, bundle]
    document.receipts = [
        {
            "gate": "build/v1",
            "status": "pass",
            "evidence": "bundle=b",
            "fingerprint": fingerprint,
            "recordedAt": "now",
        }
    ]
    errors = validate(document, Profile("t", [gate]), adapters=[adapter], workspace_root=tmp_path)
    assert any(error.invariant == 9 for error in errors)


def test_basis_scope_must_cover_declared_inputs(tmp_path: Path):
    (tmp_path / "artifact.jsonl").write_text(
        json.dumps(
            {
                "type": "command_run",
                "evidenceId": "e1",
                "gate": "build/v1",
                "check": "c",
                "exitCode": 0,
            }
        )
        + "\n",
        encoding="utf-8",
    )
    artifact = tmp_path / "artifact.jsonl"
    gate = Gate(
        "build/v1", "implementation", (Check("c", "basis", "success"),), ("source",), ("pass",)
    )
    document = Document(
        title="T",
        inputs=[{"type": "source", "path": "src/app.py"}],
        evidence=[],
        receipts=[],
        subtasks=[],
        path=tmp_path / "task.md",
    )
    evidence = {
        "kind": "basis",
        "id": "e1",
        "gate": "build/v1",
        "check": "c",
        "issuerRole": "implementation",
        "scope": [],
        "classifiedRefs": [],
        "artifactSha256": sha256(artifact),
        "basisRef": "artifact.jsonl",
    }
    document.evidence = [evidence]
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 2 and "coverage" in str(error) for error in errors)


def test_multi_record_command_artifact_is_rejected(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "task.md"
    )
    artifact = tmp_path / "run.jsonl"
    record = {
        "type": "command_run",
        "evidenceId": "e1",
        "gate": "build/v1",
        "check": "c",
        "exitCode": 0,
    }
    artifact.write_text(json.dumps(record) + "\n" + json.dumps(record) + "\n", encoding="utf-8")
    document.evidence = [
        {
            "kind": "command",
            "id": "e1",
            "gate": "build/v1",
            "check": "c",
            "issuerRole": "implementation",
            "artifactSha256": sha256(artifact),
            "commandReceipt": "run.jsonl",
        }
    ]
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 2 and "exactly one record" in str(error) for error in errors)


def test_directory_symlink_is_rejected(tmp_path: Path):
    source = tmp_path / "src"
    source.mkdir()
    target = source / "target"
    target.mkdir()
    (target / "app.py").write_text("print('ok')\n", encoding="utf-8")
    (source / "linked").symlink_to(target, target_is_directory=True)

    with pytest.raises(ValueError, match="symlink inputs are not supported"):
        fingerprint_gate(tmp_path, [{"type": "source", "path": "src"}], types=["source"])


def test_validator_rejects_unhashable_bundle_member(tmp_path: Path):
    from acdd.adapter import Adapter, CheckBinding, GateBinding

    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "t.md"
    )
    adapter = Adapter(
        "implementation",
        "implementation",
        "artifacts",
        tmp_path,
        {"build/v1": GateBinding(checks={"c": CheckBinding(argv=("/bin/true",))})},
    )
    fingerprint = fingerprint_for_gate(document, gate, tmp_path, adapter)
    document.evidence = [
        {
            "kind": "bundle",
            "id": "b",
            "gate": "build/v1",
            "issuerRole": "implementation",
            "checkEvidence": [{}],
            "inputFingerprint": fingerprint,
        }
    ]
    document.receipts = [
        {
            "gate": "build/v1",
            "status": "pass",
            "evidence": "bundle=b",
            "fingerprint": fingerprint,
            "recordedAt": "now",
        }
    ]

    errors = validate(document, Profile("t", [gate]), adapters=[adapter], workspace_root=tmp_path)
    assert any(error.invariant == 2 and "invalid bundle members" in str(error) for error in errors)


def test_missing_executable_returns_error_not_traceback(core):
    doc, profile, adapter = core
    adapter.write_text(
        adapter.read_text(encoding="utf-8").replace(
            "argv: [python3, -c, \"print('green')\"]", "argv: [nonexistent-binary-xyz]"
        ),
        encoding="utf-8",
    )
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.runtime",
            ]
        )
        == 1
    )
    artifact_text = (doc.parent / "artifacts" / "build.runtime.jsonl").read_text(encoding="utf-8")
    assert json.loads(artifact_text)["exitCode"] == 127


def test_transitive_dependency_serializes_conflict(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T",
        inputs=[{"type": "source", "path": "x.py"}],
        evidence=[],
        receipts=[
            {
                "gate": "build/v1",
                "status": "pending",
                "evidence": "pending",
                "fingerprint": "pending",
                "recordedAt": "pending",
            }
        ],
        subtasks=[
            Subtask("a", ("x.py",), (), "A"),
            Subtask("b", (), (), "B", ("a",)),
            Subtask("c", ("x.py",), (), "C", ("b",)),
        ],
        path=tmp_path / "task.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert not any(error.invariant == 6 for error in errors)


def test_unserialized_conflict_still_rejected(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T",
        inputs=[{"type": "source", "path": "x.py"}],
        evidence=[],
        receipts=[
            {
                "gate": "build/v1",
                "status": "pending",
                "evidence": "pending",
                "fingerprint": "pending",
                "recordedAt": "pending",
            }
        ],
        subtasks=[
            Subtask("a", ("x.py",), (), "A"),
            Subtask("b", (), (), "B", ("a",)),
            Subtask("c", ("x.py",), (), "C"),
        ],
        path=tmp_path / "task.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 6 for error in errors)


def test_cyclic_subtask_dependencies_rejected(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T",
        inputs=[],
        evidence=[],
        receipts=[
            {
                "gate": "build/v1",
                "status": "pending",
                "evidence": "pending",
                "fingerprint": "pending",
                "recordedAt": "pending",
            }
        ],
        subtasks=[Subtask("a", (), (), "A", ("b",)), Subtask("b", (), (), "B", ("a",))],
        path=tmp_path / "task.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 6 and "acyclic" in str(error) for error in errors)


def test_record_review_resolves_transcript_against_workspace_root(tmp_path: Path, monkeypatch):
    monkeypatch.chdir("/tmp")
    (tmp_path / "a.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: r
role: review
gates:
  review/v1:
    checks:
      independent-review: {argv: [echo]}
""",
        encoding="utf-8",
    )
    adapter = load_adapter(tmp_path / "a.yaml")
    gate = Gate(
        "review/v1",
        "review",
        (Check("independent-review", "review", "success"),),
        (),
        ("pass",),
        (),
        ("parity", "security"),
    )
    transcript = tmp_path / "rev.jsonl"
    terminal = {
        "type": "review_terminal",
        "evidenceId": "r1",
        "gate": "review/v1",
        "check": "independent-review",
        "scope": ["src/"],
        "performedChecks": ["parity", "security"],
        "verdict": "pass",
        "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
        "reviewerSessionUuid": "00000000-0000-4000-8000-000000000003",
        "reviewedSessionUuids": ["00000000-0000-4000-8000-000000000002"],
    }
    raw = _review_raw(raw="reviewer response without a machine-readable schema")
    confirmation = _review_raw(
        reviewer_uuid="00000000-0000-4000-8000-000000000003", raw="confirmation review: PASS"
    )
    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "d.md"
    )
    document.path.write_text("---\ntitle: T\n---\n## Evidence\n\n## Receipts\n", encoding="utf-8")
    from acdd.record import record_review

    def write_transcript():
        transcript.write_text(
            "\n".join(json.dumps(row) for row in (raw, confirmation, terminal)) + "\n",
            encoding="utf-8",
        )

    write_transcript()
    with pytest.raises(AcddError, match="review transcript"):
        record_review(
            document=document,
            workspace_root=tmp_path,
            gate=gate,
            check_id="independent-review",
            evidence_id="r1",
            adapter=adapter,
            transcript=Path("rev.jsonl"),
            author_uuid="00000000-0000-4000-8000-000000000001",
            reviewer_uuid="00000000-0000-4000-8000-000000000003",
            verdict="pass",
        )
    terminal["reviewedSessionUuids"].append("00000000-0000-4000-8000-000000000003")
    terminal["verdict"] = "fail"
    write_transcript()
    with pytest.raises(AcddError, match="review_terminal"):
        record_review(
            document=document,
            workspace_root=tmp_path,
            gate=gate,
            check_id="independent-review",
            evidence_id="r1",
            adapter=adapter,
            transcript=Path("rev.jsonl"),
            author_uuid="00000000-0000-4000-8000-000000000001",
            reviewer_uuid="00000000-0000-4000-8000-000000000003",
            verdict="pass",
        )
    terminal["verdict"] = "pass"
    terminal["reviewerSessionUuid"] = "00000000-0000-4000-8000-000000000004"
    write_transcript()
    with pytest.raises(AcddError, match="session UUIDs"):
        record_review(
            document=document,
            workspace_root=tmp_path,
            gate=gate,
            check_id="independent-review",
            evidence_id="r1",
            adapter=adapter,
            transcript=Path("rev.jsonl"),
            author_uuid="00000000-0000-4000-8000-000000000001",
            reviewer_uuid="00000000-0000-4000-8000-000000000003",
            verdict="pass",
        )
    terminal["reviewerSessionUuid"] = "00000000-0000-4000-8000-000000000003"
    write_transcript()
    payload = record_review(
        document=document,
        workspace_root=tmp_path,
        gate=gate,
        check_id="independent-review",
        evidence_id="r1",
        adapter=adapter,
        transcript=Path("rev.jsonl"),
        author_uuid="00000000-0000-4000-8000-000000000001",
        reviewer_uuid="00000000-0000-4000-8000-000000000003",
        verdict="pass",
    )
    assert payload["transcriptRef"] == "rev.jsonl"
    assert payload["artifactSha256"] == sha256(transcript)


def test_non_dict_inputs_entry_is_a_clean_error(tmp_path: Path):
    (tmp_path / "d.md").write_text(
        """\
---
title: T
---
## Inputs
```yaml
paths:
  - just-a-string
```
## Receipts
""",
        encoding="utf-8",
    )
    with pytest.raises(AcddError, match="invalid Inputs entry"):
        load_document(tmp_path / "d.md")


def test_adapter_cwd_escape_is_rejected(core):
    doc, profile, adapter = core
    adapter.write_text(
        adapter.read_text(encoding="utf-8").replace("cwd: .", "cwd: /etc"), encoding="utf-8"
    )
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.runtime",
            ]
        )
        == 2
    )


def test_adapter_timeout_seconds_is_validated(tmp_path: Path):
    path = tmp_path / "a.yaml"
    path.write_text(
        """\
apiVersion: acdd/adapter/v1
id: x
role: implementation
gates:
  build/v1:
    checks:
      c: {argv: [/bin/true], timeoutSeconds: 0}
""",
        encoding="utf-8",
    )
    with pytest.raises(AdapterError, match="timeoutSeconds"):
        load_adapter(path)
    path.write_text(
        path.read_text(encoding="utf-8").replace("timeoutSeconds: 0", "timeoutSeconds: 5"),
        encoding="utf-8",
    )
    assert load_adapter(path).gates["build/v1"].checks["c"].timeout_seconds == 5


def test_adapter_rejects_non_string_artifact_dir(tmp_path: Path):
    path = tmp_path / "a.yaml"
    path.write_text(
        """
apiVersion: acdd/adapter/v1
id: x
role: implementation
artifactDir: []
gates: {}
""",
        encoding="utf-8",
    )

    with pytest.raises(AdapterError, match="artifactDir"):
        load_adapter(path)


def test_partial_receipt_must_be_fully_pending(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    document = Document(
        title="T",
        inputs=[],
        evidence=[],
        receipts=[
            {
                "gate": "build/v1",
                "status": "partial",
                "evidence": "garbage",
                "fingerprint": "garbage",
                "recordedAt": "never",
            }
        ],
        subtasks=[],
        path=tmp_path / "task.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 1 for error in errors)


def test_document_profile_mismatch_is_rejected(core):
    doc, profile, adapter = core
    doc.write_text(
        doc.read_text(encoding="utf-8").replace(
            "planning_profile: test/v1", "planning_profile: other/v1"
        ),
        encoding="utf-8",
    )
    errors = validate(
        load_document(doc),
        load_profile(profile),
        adapters=[load_adapter(adapter)],
        workspace_root=doc.parent,
    )
    assert any(error.invariant == 1 and "declares" in str(error) for error in errors)


def test_plan_edit_does_not_stale_receipt(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
            ]
        )
        == 0
    )
    doc.write_text(
        doc.read_text(encoding="utf-8").replace(
            "subtasks: []",
            """subtasks:
  - id: docs
    writes: [src/app.py]
    reads: []
    acceptance: guide written""",
        ),
        encoding="utf-8",
    )
    errors = validate(
        load_document(doc),
        load_profile(profile),
        adapters=[load_adapter(adapter)],
        workspace_root=doc.parent,
    )
    assert not any(error.invariant == 4 for error in errors)


def test_unrelated_input_type_does_not_stale_receipt(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
            ]
        )
        == 0
    )
    doc.write_text(
        doc.read_text(encoding="utf-8").replace(
            """paths:
  - {type: source, path: src/app.py}""",
            """paths:
  - {type: source, path: src/app.py}
  - {type: docs, path: docs/guide.md}""",
        ),
        encoding="utf-8",
    )
    errors = validate(
        load_document(doc),
        load_profile(profile),
        adapters=[load_adapter(adapter)],
        workspace_root=doc.parent,
    )
    assert not any(error.invariant == 4 for error in errors)


def test_note_column_rules(tmp_path: Path):
    gate = Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",))
    profile = Profile("t", [gate])

    def errors_for(status: str, note: str | None):
        cells = ["build/v1", status, "pending", "pending", "pending"] + (
            [note] if note is not None else []
        )
        document = Document(
            title="T",
            inputs=[],
            evidence=[],
            receipts=[
                dict(
                    zip(("gate", "status", "evidence", "fingerprint", "recordedAt", "note"), cells)
                )
            ],
            subtasks=[],
            path=tmp_path / "t.md",
        )
        return validate(document, profile, workspace_root=tmp_path)

    assert not any(e.invariant == 1 for e in errors_for("blocked", "waiting on infra"))
    assert not any(e.invariant == 1 for e in errors_for("partial", None))
    assert any(e.invariant == 1 for e in errors_for("pending", "not allowed"))
    assert any(e.invariant == 1 for e in errors_for("blocked", ""))


def test_acdd_discovery_loads_adapters(core):
    doc, profile, adapter = core
    discovered = doc.parent / "sub" / ".acdd"
    discovered.mkdir(parents=True)
    (discovered / "impl.yaml").write_text(adapter.read_text(encoding="utf-8"), encoding="utf-8")
    args = [str(doc), str(profile), "--workspace-root", str(doc.parent)]
    assert (
        main(
            [
                "record",
                *args,
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    assert main(["finalize", *args, "--gate", "build/v1", "--id", "build.bundle"]) == 0
    assert main(["validate", *args]) == 0


def test_acdd_discovery_rejects_duplicate_roles(core):
    doc, profile, adapter = core
    for name in ("one", "two"):
        directory = doc.parent / name / ".acdd"
        directory.mkdir(parents=True)
        (directory / "impl.yaml").write_text(adapter.read_text(encoding="utf-8"), encoding="utf-8")
    assert main(["validate", str(doc), str(profile), "--workspace-root", str(doc.parent)]) == 2


def test_adapter_outside_workspace_is_rejected(core):
    doc, profile, _adapter = core
    assert (
        main(
            [
                "validate",
                str(doc),
                str(profile),
                "--workspace-root",
                str(doc.parent),
                "--adapter",
                "implementation=/etc/passwd",
            ]
        )
        == 2
    )


def test_discovery_prunes_vendor_dirs(core):
    doc, _profile, adapter = core
    stray = doc.parent / ".venv" / "lib" / ".acdd"
    stray.mkdir(parents=True)
    (stray / "stray.yaml").write_text(adapter.read_text(encoding="utf-8"), encoding="utf-8")
    from acdd.cli import _discover

    assert _discover(doc.parent) == []
    nested = doc.parent / "pkg" / ".acdd"
    nested.mkdir(parents=True)
    (nested / "impl.yaml").write_text(adapter.read_text(encoding="utf-8"), encoding="utf-8")
    assert _discover(doc.parent) == [nested / "impl.yaml"]


def test_overflow_receipt_row_is_a_shape_error(core):
    doc, profile, adapter = core
    doc.write_text(
        doc.read_text(encoding="utf-8").replace(
            "| build/v1 | pending | pending | pending | pending |",
            "| build/v1 | pending | pending | pending | pending | note | junk |",
        ),
        encoding="utf-8",
    )
    errors = validate(
        load_document(doc),
        load_profile(profile),
        adapters=[load_adapter(adapter)],
        workspace_root=doc.parent,
    )
    assert any(error.invariant == 1 for error in errors)


def test_review_dimensions_required_in_transcript(tmp_path: Path):
    (tmp_path / "p.yaml").write_text(
        """\
apiVersion: acdd/profile/v1
kind: profile
id: t
gates:
  - id: review/v1
    owner: review
    checks:
      - {id: independent-review, evidenceKind: review}
    invalidatesOn: []
    reviewDimensions: [parity, security]
    terminals: [pass]
""",
        encoding="utf-8",
    )
    (tmp_path / "a.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: r
role: review
gates:
  review/v1:
    checks:
      independent-review: {argv: [pi, --print, "{prompt}"]}
""",
        encoding="utf-8",
    )
    (tmp_path / "d.md").write_text(
        "---\ntitle: T\n---\n## Evidence\n\n## Receipts\n", encoding="utf-8"
    )
    adapter = load_adapter(tmp_path / "a.yaml")
    gate = Gate(
        "review/v1",
        "review",
        (Check("independent-review", "review", "success"),),
        (),
        ("pass",),
        (),
        ("parity", "security"),
    )
    transcript = tmp_path / "rev.jsonl"
    raw = _review_raw()
    terminal = {
        "type": "review_terminal",
        "evidenceId": "r1",
        "gate": "review/v1",
        "check": "independent-review",
        "scope": ["src/"],
        "performedChecks": ["parity"],
        "verdict": "pass",
        "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
        "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
        "reviewedSessionUuids": ["00000000-0000-4000-8000-000000000002"],
    }
    transcript.write_text(
        "\n".join(json.dumps(row) for row in (raw, terminal)) + "\n", encoding="utf-8"
    )
    from acdd.record import record_review

    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "d.md"
    )
    with pytest.raises(AcddError, match="covering"):
        record_review(
            document=document,
            workspace_root=tmp_path,
            gate=gate,
            check_id="independent-review",
            evidence_id="r1",
            adapter=adapter,
            transcript=transcript,
            author_uuid="00000000-0000-4000-8000-000000000001",
            reviewer_uuid="00000000-0000-4000-8000-000000000002",
            verdict="pass",
        )
    terminal["performedChecks"] = ["parity", "security", "performance"]
    transcript.write_text(
        "\n".join(json.dumps(row) for row in (raw, terminal)) + "\n", encoding="utf-8"
    )
    record_review(
        document=document,
        workspace_root=tmp_path,
        gate=gate,
        check_id="independent-review",
        evidence_id="r1",
        adapter=adapter,
        transcript=transcript,
        author_uuid="00000000-0000-4000-8000-000000000001",
        reviewer_uuid="00000000-0000-4000-8000-000000000002",
        verdict="pass",
    )


def test_discovery_rejects_unknown_check_on_active_gate(core):
    doc, profile, _adapter = core
    (doc.parent / "side" / ".acdd").mkdir(parents=True)
    (doc.parent / "side" / ".acdd" / "x.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: stray
role: implementation
gates:
  build/v1:
    checks:
      c: {argv: [/bin/true]}
  mystery/v1:
    checks:
      c: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )
    buf = io.StringIO()
    original, sys.stderr = sys.stderr, buf
    try:
        rc = main(["validate", str(doc), str(profile), "--workspace-root", str(doc.parent)])
    finally:
        sys.stderr = original
    assert rc == 1
    err = buf.getvalue()
    assert "invariant 5" in err
    assert "unknown check" in err


def test_discovery_ignores_adapters_for_inactive_profile_roles(core):
    doc, profile, adapter = core
    (doc.parent / "plan" / ".acdd").mkdir(parents=True)
    (doc.parent / "plan" / ".acdd" / "plan.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: plan
role: plan
gates:
  decompose/v1:
    checks:
      matrix: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )

    assert (
        main(
            [
                "validate",
                str(doc),
                str(profile),
                "--workspace-root",
                str(doc.parent),
                "--adapter",
                f"implementation={adapter}",
            ]
        )
        == 0
    )


def test_profile_rejects_review_dimensions_without_review_check(tmp_path: Path):
    path = tmp_path / "p.yaml"
    path.write_text(
        """\
apiVersion: acdd/profile/v1
kind: profile
id: t
gates:
  - id: build/v1
    owner: implementation
    checks:
      - {id: c, evidenceKind: command}
    invalidatesOn: []
    reviewDimensions: [parity]
    terminals: [pass]
""",
        encoding="utf-8",
    )
    with pytest.raises(AcddError, match="reviewDimensions"):
        load_profile(path)


def test_review_transcript_non_string_elements_rejected(tmp_path: Path):
    gate = Gate(
        "review/v1",
        "review",
        (Check("independent-review", "review", "success"),),
        (),
        ("pass",),
        (),
        ("parity", "security"),
    )
    artifact = tmp_path / "rev.jsonl"
    terminal = {
        "type": "review_terminal",
        "evidenceId": "r1",
        "gate": "review/v1",
        "check": "independent-review",
        "scope": ["src/"],
        "performedChecks": [42, 42, "parity"],
        "verdict": "pass",
        "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
        "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
        "reviewedSessionUuids": ["00000000-0000-4000-8000-000000000002"],
    }
    artifact.write_text(
        "\n".join(json.dumps(row) for row in (_review_raw(), terminal)) + "\n", encoding="utf-8"
    )
    document = Document(
        title="T",
        inputs=[],
        evidence=[
            {
                "kind": "review",
                "id": "r1",
                "gate": "review/v1",
                "check": "independent-review",
                "issuerRole": "review",
                "transcriptRef": "rev.jsonl",
                "artifactSha256": sha256(artifact),
                "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
                "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
                "verdict": "pass",
            }
        ],
        receipts=[],
        subtasks=[],
        path=tmp_path / "t.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 11 for error in errors)
    artifact.write_text("not JSON\n", encoding="utf-8")
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 2 for error in errors)


def test_discovery_rejects_symlinked_adapter_files(tmp_path: Path):
    import os

    real = tmp_path / "real.yaml"
    real.write_text("apiVersion: acdd/adapter/v1\nid: x\nrole: task\ngates: {}", encoding="utf-8")
    (tmp_path / ".acdd").mkdir()
    lnk = tmp_path / ".acdd" / "x.yaml"
    os.symlink(real, lnk)
    from acdd.cli import _discover

    assert _discover(tmp_path) == []


def test_failed_record_allows_retry_same_id(core):
    doc, profile, adapter = core
    adapter.write_text(
        adapter.read_text(encoding="utf-8").replace(
            "argv: [python3, -c, \"print('green')\"]",
            'argv: [python3, -c, "import sys; sys.exit(1)"]',
        ),
        encoding="utf-8",
    )
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.retry",
            ]
        )
        == 1
    )
    assert (doc.parent / "artifacts" / "build.retry.jsonl").is_file()
    adapter.write_text(
        adapter.read_text(encoding="utf-8").replace(
            'argv: [python3, -c, "import sys; sys.exit(1)"]',
            "argv: [python3, -c, \"print('green')\"]",
        ),
        encoding="utf-8",
    )
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.retry",
            ]
        )
        == 0
    )


def test_expected_failure_rejects_reserved_exit_codes(tmp_path: Path):
    artifact = tmp_path / "run.jsonl"
    artifact.write_text(
        json.dumps(
            {
                "type": "command_run",
                "evidenceId": "e1",
                "gate": "build/v1",
                "check": "c",
                "exitCode": 124,
            }
        )
        + "\n",
        encoding="utf-8",
    )
    gate = Gate(
        "build/v1", "implementation", (Check("c", "command", "expected-failure"),), (), ("pass",)
    )
    document = Document(
        title="T",
        inputs=[],
        evidence=[
            {
                "kind": "command",
                "id": "e1",
                "gate": "build/v1",
                "check": "c",
                "issuerRole": "implementation",
                "artifactSha256": sha256(artifact),
                "commandReceipt": "run.jsonl",
            }
        ],
        receipts=[],
        subtasks=[],
        path=tmp_path / "t.md",
    )
    errors = validate(document, Profile("t", [gate]), workspace_root=tmp_path)
    assert any(error.invariant == 9 for error in errors)


def test_record_review_rejects_non_terminal_type(tmp_path: Path):
    (tmp_path / "a.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: r
role: review
gates:
  review/v1:
    checks:
      independent-review: {argv: [echo]}
""",
        encoding="utf-8",
    )
    adapter = load_adapter(tmp_path / "a.yaml")
    gate = Gate(
        "review/v1",
        "review",
        (Check("independent-review", "review", "success"),),
        (),
        ("pass",),
        (),
        ("parity",),
    )
    transcript = tmp_path / "rev.jsonl"
    transcript.write_text(
        json.dumps(
            {
                "type": "command_run",
                "evidenceId": "r1",
                "gate": "review/v1",
                "check": "independent-review",
                "scope": ["src/"],
                "performedChecks": ["parity"],
                "exitCode": 0,
                "verdict": "pass",
                "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
                "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
            }
        )
        + "\n",
        encoding="utf-8",
    )
    document = Document(
        title="T", inputs=[], evidence=[], receipts=[], subtasks=[], path=tmp_path / "d.md"
    )
    document.path.write_text("---\ntitle: T\n---\n## Evidence\n\n## Receipts\n", encoding="utf-8")
    from acdd.record import record_review

    with pytest.raises(AcddError, match="review_terminal"):
        record_review(
            document=document,
            workspace_root=tmp_path,
            gate=gate,
            check_id="independent-review",
            evidence_id="r1",
            adapter=adapter,
            transcript=transcript,
            author_uuid="00000000-0000-4000-8000-000000000001",
            reviewer_uuid="00000000-0000-4000-8000-000000000002",
            verdict="pass",
        )


def test_terminal_gate_requires_prior_gates(tmp_path: Path):
    profile = Profile(
        "t",
        [
            Gate("design/v1", "task", (Check("c", "command", "success"),), (), ("pass",)),
            Gate("build/v1", "implementation", (Check("c", "command", "success"),), (), ("pass",)),
        ],
    )
    document = Document(
        title="T",
        inputs=[],
        evidence=[],
        receipts=[
            {
                "gate": "design/v1",
                "status": "pending",
                "evidence": "pending",
                "fingerprint": "pending",
                "recordedAt": "pending",
            },
            {
                "gate": "build/v1",
                "status": "pass",
                "evidence": "bundle=b",
                "fingerprint": "fp",
                "recordedAt": "now",
            },
        ],
        subtasks=[],
        path=tmp_path / "t.md",
    )
    errors = validate(document, profile, workspace_root=tmp_path)
    assert any(error.invariant == 1 and "prior gate" in str(error) for error in errors)


def test_duplicate_check_evidence_rejected_on_record(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.one",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.two",
            ]
        )
        == 2
    )


def test_inapplicable_rejects_check_evidence(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
                "--status",
                "inapplicable",
                "--reason-code",
                "build.no-runnable-source",
            ]
        )
        == 2
    )


def test_inapplicable_with_empty_bundle_passes(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.inapp",
                "--status",
                "inapplicable",
                "--reason-code",
                "build.no-runnable-source",
            ]
        )
        == 0
    )
    assert (
        validate(
            load_document(doc),
            load_profile(profile),
            adapters=[load_adapter(adapter)],
            workspace_root=doc.parent,
        )
        == []
    )


def test_profile_rejects_scalar_invalidates_on(tmp_path: Path):
    profile = tmp_path / "profile.yaml"
    profile.write_text(
        """
apiVersion: acdd/profile/v1
kind: profile
id: t
gates:
  - id: build/v1
    owner: implementation
    checks: [{id: c, evidenceKind: command}]
    invalidatesOn: source
    terminals: [pass]
""",
        encoding="utf-8",
    )
    with pytest.raises(AcddError, match="invalidatesOn"):
        load_profile(profile)


def test_cli_reports_malformed_profile_yaml(tmp_path: Path, capsys):
    profile = tmp_path / "profile.yaml"
    profile.write_text("gates: [", encoding="utf-8")
    assert main(["validate", str(tmp_path / "missing.md"), str(profile)]) == 2
    assert "ACDD ERROR:" in capsys.readouterr().err


def test_finalize_rejects_missing_predecessor_receipt(core):
    doc, profile, adapter = core
    profile.write_text(
        """
apiVersion: acdd/profile/v1
kind: profile
id: test/v1
gates:
  - id: design/v1
    owner: task
    checks: [{id: c, evidenceKind: command}]
    invalidatesOn: []
    terminals: [pass]
  - id: build/v1
    owner: implementation
    checks: [{id: runtime-and-integration, evidenceKind: command}]
    invalidatesOn: [source]
    terminals: [pass, inapplicable]
    inapplicableReasonCodes: [build.no-runnable-source]
""",
        encoding="utf-8",
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
                "--status",
                "inapplicable",
                "--reason-code",
                "build.no-runnable-source",
            ]
        )
        == 2
    )
    assert "build.bundle" not in doc.read_text(encoding="utf-8")


def test_finalize_rejects_tampered_current_evidence(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    (doc.parent / "artifacts" / "build.check.jsonl").write_text("tampered\n", encoding="utf-8")
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
            ]
        )
        == 2
    )
    assert "build.bundle" not in doc.read_text(encoding="utf-8")


def test_finalize_rejects_terminal_gate(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.check",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.bundle",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.second",
            ]
        )
        == 2
    )


def test_finalize_replaces_stale_terminal_gate(core):
    doc, profile, adapter = core
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.before",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.before.bundle",
            ]
        )
        == 0
    )
    (doc.parent / "src" / "app.py").write_text("print('fixed')\n", encoding="utf-8")
    assert (
        main(
            [
                "record",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--check",
                "runtime-and-integration",
                "--id",
                "build.after",
            ]
        )
        == 0
    )
    assert (
        main(
            [
                "finalize",
                *_args(doc, profile, adapter),
                "--gate",
                "build/v1",
                "--id",
                "build.after.bundle",
            ]
        )
        == 0
    )
    assert (
        validate(
            load_document(doc),
            load_profile(profile),
            adapters=[load_adapter(adapter)],
            workspace_root=doc.parent,
        )
        == []
    )


def test_command_outcome_rejects_boolean_exit_code():
    assert not command_outcome_ok(
        Check("c", "command", "success"), {"type": "command_run", "exitCode": False}
    )



from __future__ import annotations

import hashlib
from dataclasses import dataclass, field
from pathlib import Path

import yaml


class AdapterError(ValueError): ...


@dataclass(frozen=True)
class CheckBinding:
    cwd: str = "."
    argv: tuple[str, ...] = ()
    prompt_append: str | None = None
    timeout_seconds: int = 300


@dataclass(frozen=True)
class GateBinding:
    checks: dict[str, CheckBinding] = field(default_factory=dict)
    contract_sections: tuple[str, ...] = ()


@dataclass(frozen=True)
class Adapter:
    id: str
    role: str
    artifact_dir: str
    base_dir: Path
    gates: dict[str, GateBinding] = field(default_factory=dict)

    def resolve(self, value: str) -> Path:
        return (self.base_dir / value).resolve()

    def prompt_digest(self, binding: CheckBinding) -> str | None:
        if binding.prompt_append is None:
            return None
        path = self.resolve(binding.prompt_append)
        try:
            path.relative_to(self.base_dir)
        except ValueError as exc:
            raise AdapterError("promptAppend escapes adapter directory") from exc
        if not path.is_file():
            raise AdapterError(f"promptAppend is missing: {binding.prompt_append!r}")
        return f"sha256:{hashlib.sha256(path.read_bytes()).hexdigest()}"


def _mapping(value: object, label: str) -> dict:
    if not isinstance(value, dict):
        raise AdapterError(f"{label} must be a mapping")
    return value


def _parse_check(gate_id: str, check_id: str, raw_check: object) -> CheckBinding:
    check = _mapping(raw_check, f"binding {gate_id}.{check_id}")
    if set(check) - {"cwd", "argv", "promptAppend", "timeoutSeconds"}:
        raise AdapterError(f"unsupported binding fields for {gate_id}.{check_id}")
    argv = check.get("argv")
    if (
        not isinstance(argv, list)
        or not argv
        or not all(isinstance(arg, str) and arg for arg in argv)
    ):
        raise AdapterError(f"binding {gate_id}.{check_id} requires non-empty string argv")
    cwd = check.get("cwd", ".")
    if not isinstance(cwd, str):
        raise AdapterError(f"binding {gate_id}.{check_id} cwd must be a string")
    prompt_append = check.get("promptAppend")
    if prompt_append is not None and (not isinstance(prompt_append, str) or not prompt_append):
        raise AdapterError(f"binding {gate_id}.{check_id} promptAppend must be a non-empty string")
    timeout = check.get("timeoutSeconds", 300)
    if not isinstance(timeout, int) or isinstance(timeout, bool) or timeout < 1:
        raise AdapterError(
            f"binding {gate_id}.{check_id} timeoutSeconds must be a positive integer"
        )
    return CheckBinding(
        cwd=cwd, argv=tuple(argv), prompt_append=prompt_append, timeout_seconds=timeout
    )


def load_adapter(path: Path) -> Adapter:
    path = path.resolve()
    data = _mapping(yaml.safe_load(path.read_text(encoding="utf-8")) or {}, "adapter")
    unknown = set(data) - {"apiVersion", "id", "role", "artifactDir", "gates"}
    if unknown:
        raise AdapterError(f"unsupported adapter fields: {sorted(unknown)}")
    if data.get("apiVersion") != "acdd/adapter/v1":
        raise AdapterError("adapter apiVersion must be acdd/adapter/v1")
    artifact_dir = data.get("artifactDir", "artifacts")
    if not isinstance(artifact_dir, str):
        raise AdapterError("adapter artifactDir must be a string")
    for key in ("id", "role"):
        if not isinstance(data.get(key), str) or not data[key]:
            raise AdapterError(f"adapter {key} is required")
    gates: dict[str, GateBinding] = {}
    for gate_id, raw_gate in _mapping(data.get("gates") or {}, "adapter gates").items():
        gate = _mapping(raw_gate, f"adapter gate {gate_id}")
        unknown_gate = set(gate) - {"checks", "contractSections"}
        if unknown_gate or "checks" not in gate:
            raise AdapterError(
                f"adapter gate {gate_id} must contain checks and optional contractSections"
            )
        sections = gate.get("contractSections")
        if sections is None:
            contract_sections: tuple[str, ...] = ()
        elif (
            not isinstance(sections, list)
            or not sections
            or not all(isinstance(item, str) and item.strip() for item in sections)
        ):
            raise AdapterError(
                f"adapter gate {gate_id} contractSections must be a non-empty list of headings"
            )
        else:
            contract_sections = tuple(item.strip() for item in sections)
        checks = {
            cid: _parse_check(gate_id, cid, raw)
            for cid, raw in _mapping(gate["checks"], f"checks for {gate_id}").items()
        }
        gates[gate_id] = GateBinding(checks=checks, contract_sections=contract_sections)
    return Adapter(
        id=data["id"],
        role=data["role"],
        artifact_dir=data.get("artifactDir", "artifacts"),
        base_dir=path.parent,
        gates=gates,
    )


def index_adapters(adapters: list[Adapter]) -> dict[str, Adapter]:
    indexed: dict[str, Adapter] = {}
    for adapter in adapters:
        if adapter.role in indexed:
            raise AdapterError(f"duplicate adapter role {adapter.role!r}")
        indexed[adapter.role] = adapter
    return indexed



"""Contract authority digest, material classify, write-set, pre-contract."""

from __future__ import annotations

import json
import subprocess
from pathlib import Path

import pytest

from acdd.adapter import Adapter, CheckBinding, GateBinding
from acdd.authority import (
    assert_changed_paths_allowed,
    assert_precontract_clean,
    assert_writes_not_shrunk,
    authority_digest,
    classify_contract_change,
)
from acdd.model import AcddError, Check, Gate, Profile, Subtask, load_document
from acdd.record import (
    finalize_gate,
    record_check,
    record_review,
    record_subtask_contract,
    reopen_gate,
)
from acdd.validate import validate


def _git_init(root: Path) -> None:
    subprocess.run(["git", "init"], cwd=root, check=True, capture_output=True)
    subprocess.run(
        ["git", "config", "user.email", "t@example.com"], cwd=root, check=True, capture_output=True
    )
    subprocess.run(["git", "config", "user.name", "t"], cwd=root, check=True, capture_output=True)
    subprocess.run(["git", "add", "-A"], cwd=root, check=True, capture_output=True)
    subprocess.run(["git", "commit", "-m", "init"], cwd=root, check=True, capture_output=True)


def _doc(tmp_path: Path, *, with_review: bool = False) -> tuple[Path, Profile, list[Adapter]]:
    (tmp_path / "src").mkdir()
    (tmp_path / "src" / "app.py").write_text("print('ok')\n", encoding="utf-8")
    (tmp_path / "src" / "second.py").write_text("print('second')\n", encoding="utf-8")
    document_path = tmp_path / "task.md"
    document_path.write_text(
        """\
---
title: T
planning_profile: task/v1
---
## Plan
```yaml
subtasks:
  - id: first
    writes: [src/app.py]
    reads: []
    acceptance: first behavior
```
## Inputs
```yaml
paths:
  - {type: source, path: src}
```
## Evidence

## Receipts
| gate | status | evidence | fingerprint | recordedAt |
| --- | --- | --- | --- | --- |
| contract/v1 | pending | pending | pending | pending |
| build/v1 | pending | pending | pending | pending |
""",
        encoding="utf-8",
    )
    checks: tuple[Check, ...] = ()
    gates_binding: dict[str, GateBinding] = {"contract/v1": GateBinding()}
    verify_binding: dict[str, GateBinding] = {}
    if with_review:
        checks = (Check("contract-verify", "review", "success", "contract-verify"),)
        gates_binding = {"contract/v1": GateBinding()}
        verify_binding = {
            "contract/v1": GateBinding(
                checks={"contract-verify": CheckBinding(argv=("true",), cwd=".")}
            )
        }
    contract = Gate(
        "contract/v1",
        "task",
        checks,
        ("source",),
        ("pass",),
        review_dimensions=("completeness",) if with_review else (),
    )
    build = Gate(
        "build/v1",
        "implementation",
        (Check("runtime-and-integration", "command", "success"),),
        ("source",),
        ("pass",),
    )
    adapters = [
        Adapter("task", "task", "artifacts", tmp_path, gates_binding),
        Adapter(
            "implementation",
            "implementation",
            "artifacts",
            tmp_path,
            {
                "build/v1": GateBinding(
                    checks={"runtime-and-integration": CheckBinding(argv=("true",), cwd=".")}
                )
            },
        ),
    ]
    if with_review:
        adapters.append(Adapter("verify", "contract-verify", "artifacts", tmp_path, verify_binding))
    return document_path, Profile("task/v1", [contract, build]), adapters


def _transcript(path: Path, *, evidence_id: str, dimensions: list[str]) -> None:
    author = "11111111-1111-4111-8111-111111111111"
    inspector = "33333333-3333-4333-8333-333333333333"
    confirmer = "22222222-2222-4222-8222-222222222222"
    rows = [
        {"type": "review_raw", "reviewerSessionUuid": inspector, "raw": "partition ok"},
        {"type": "review_raw", "reviewerSessionUuid": confirmer, "raw": "confirm pass"},
        {
            "type": "review_terminal",
            "evidenceId": evidence_id,
            "gate": "contract/v1",
            "check": "contract-verify",
            "verdict": "pass",
            "authorSessionUuid": author,
            "reviewerSessionUuid": confirmer,
            "scope": ["contract"],
            "reviewedSessionUuids": [inspector, confirmer],
            "performedChecks": dimensions,
        },
    ]
    path.write_text("".join(json.dumps(row) + "\n" for row in rows), encoding="utf-8")


def test_classify_material_for_supersede_and_overlap():
    first = Subtask("first", ("src/a.py",), (), "a")
    assert (
        classify_contract_change(
            Subtask("fix", ("src/b.py",), (), "b", supersedes="first"),
            contracted_active=(first,),
        )
        == "material"
    )
    assert (
        classify_contract_change(
            Subtask("overlap", ("src/a.py",), (), "o"), contracted_active=(first,)
        )
        == "material"
    )
    assert (
        classify_contract_change(Subtask("add", ("src/b.py",), (), "n"), contracted_active=(first,))
        == "additive"
    )


def test_classify_repair_id_is_material():
    first = Subtask("first", ("src/a.py",), (), "a")
    assert (
        classify_contract_change(
            Subtask("wave1-repair", ("src/b.py",), (), "r", depends_on=("first",)),
            contracted_active=(first,),
        )
        == "material"
    )
    assert (
        classify_contract_change(
            Subtask("finding-fix", ("src/c.py",), (), "f"),
            contracted_active=(first,),
        )
        == "material"
    )


def test_write_set_rejects_escape():
    with pytest.raises(AcddError, match="outside active subtask writes"):
        assert_changed_paths_allowed(["src/other.py"], allowed_writes=["src/app.py"])
    assert_changed_paths_allowed(["src/app.py"], allowed_writes=["src/app.py"])
    assert_changed_paths_allowed(["src/pkg/x.py"], allowed_writes=["src/pkg"])


def test_anti_shrink_rejects_missing_prior_write():
    with pytest.raises(AcddError, match="shrank writes"):
        assert_writes_not_shrunk(
            ["src/app.py", "src/second.py"],
            ["src/app.py"],
            allow_scope_reduction=False,
        )
    assert_writes_not_shrunk(
        ["src/app.py", "src/second.py"],
        ["src/app.py"],
        allow_scope_reduction=True,
    )


def test_precontract_blocks_source_dirty(tmp_path: Path):
    document_path, _, _ = _doc(tmp_path)
    document = load_document(document_path)
    with pytest.raises(AcddError, match="before contract/v1 pass"):
        assert_precontract_clean(
            document_path=document.path,
            inputs=document.inputs,
            receipts=document.receipts,
            workspace=tmp_path,
            dirty_paths=["src/app.py"],
        )
    assert_precontract_clean(
        document_path=document.path,
        inputs=document.inputs,
        receipts=document.receipts,
        workspace=tmp_path,
        dirty_paths=["tests/test_proof.py"],
    )


def test_supersede_appends_via_contract_subtask(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path)
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    document_path.write_text(
        document_path.read_text(encoding="utf-8").replace(
            "acceptance: first behavior\n",
            "acceptance: first behavior\n"
            "  - id: replacement\n    writes: [src/app.py]\n    reads: []\n"
            "    acceptance: replacement\n    supersedes: first\n",
        ),
        encoding="utf-8",
    )
    part = record_subtask_contract(
        document=load_document(document_path),
        profile=profile,
        workspace_root=tmp_path,
        adapter=adapters[0],
        adapters=adapters,
        subtask_id="replacement",
        evidence_id="contract.replacement",
    )
    assert part["subtask"] == "replacement"
    assert part["supersedes"] == "first"


def test_reopen_forbidden_after_freeze(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path)
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    with pytest.raises(AcddError, match="reopen of contract/v1 is forbidden"):
        reopen_gate(
            document=load_document(document_path),
            gate=profile.gates[0],
            workspace_root=tmp_path,
        )


def test_replacement_append_rejects_write_shrink(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path)
    document_path.write_text(
        document_path.read_text(encoding="utf-8").replace(
            "writes: [src/app.py]",
            "writes: [src/app.py, src/second.py]",
        ),
        encoding="utf-8",
    )
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    document_path.write_text(
        document_path.read_text(encoding="utf-8").replace(
            "acceptance: first behavior\n",
            "acceptance: first behavior\n"
            "  - id: replacement\n    writes: [src/app.py]\n    reads: []\n"
            "    acceptance: replacement\n    supersedes: first\n",
        ),
        encoding="utf-8",
    )
    with pytest.raises(AcddError, match="shrank writes"):
        record_subtask_contract(
            document=load_document(document_path),
            profile=profile,
            workspace_root=tmp_path,
            adapter=adapters[0],
            adapters=adapters,
            subtask_id="replacement",
            evidence_id="contract.replacement",
        )
    part = record_subtask_contract(
        document=load_document(document_path),
        profile=profile,
        workspace_root=tmp_path,
        adapter=adapters[0],
        adapters=adapters,
        subtask_id="replacement",
        evidence_id="contract.replacement",
        allow_scope_reduction=True,
    )
    assert part["subtask"] == "replacement"


def test_build_requires_changed_or_git(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path)
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    with pytest.raises(AcddError, match="git worktree or explicit --changed"):
        record_check(
            document=load_document(document_path),
            workspace_root=tmp_path,
            gate=profile.gates[1],
            check_id="runtime-and-integration",
            evidence_id="build.1",
            adapter=adapters[1],
            adapters=adapters,
        )
    with pytest.raises(AcddError, match="outside active subtask writes"):
        record_check(
            document=load_document(document_path),
            workspace_root=tmp_path,
            gate=profile.gates[1],
            check_id="runtime-and-integration",
            evidence_id="build.2",
            adapter=adapters[1],
            adapters=adapters,
            changed_paths=["src/second.py"],
        )
    payload, ok = record_check(
        document=load_document(document_path),
        workspace_root=tmp_path,
        gate=profile.gates[1],
        check_id="runtime-and-integration",
        evidence_id="build.3",
        adapter=adapters[1],
        adapters=adapters,
        changed_paths=["src/app.py"],
    )
    assert ok and payload is not None


def test_build_git_write_set(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path)
    _git_init(tmp_path)
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    (tmp_path / "src" / "second.py").write_text("print('dirty')\n", encoding="utf-8")
    with pytest.raises(AcddError, match="outside active subtask writes"):
        record_check(
            document=load_document(document_path),
            workspace_root=tmp_path,
            gate=profile.gates[1],
            check_id="runtime-and-integration",
            evidence_id="build.git",
            adapter=adapters[1],
            adapters=adapters,
        )


def test_authority_digest_requires_matching_verify_after_additive(tmp_path: Path):
    document_path, profile, adapters = _doc(tmp_path, with_review=True)
    transcript = tmp_path / "artifacts" / "verify.jsonl"
    transcript.parent.mkdir(parents=True, exist_ok=True)
    _transcript(transcript, evidence_id="contract.verify", dimensions=["completeness"])
    record_review(
        document=load_document(document_path),
        workspace_root=tmp_path,
        gate=profile.gates[0],
        check_id="contract-verify",
        evidence_id="contract.verify",
        adapter=adapters[2],
        transcript=transcript,
        author_uuid="11111111-1111-4111-8111-111111111111",
        reviewer_uuid="22222222-2222-4222-8222-222222222222",
        verdict="pass",
        adapters=adapters,
    )
    digest_before = authority_digest(load_document(document_path).subtasks)
    evidence = load_document(document_path).evidence[0]
    assert evidence["authorityDigest"] == digest_before
    finalize_gate(
        document=load_document(document_path),
        profile=profile,
        adapters=adapters,
        workspace_root=tmp_path,
        gate=profile.gates[0],
        evidence_id="contract.bundle",
        adapter=adapters[0],
    )
    assert (
        validate(load_document(document_path), profile, adapters=adapters, workspace_root=tmp_path)
        == []
    )

    document_path.write_text(
        document_path.read_text(encoding="utf-8").replace(
            "acceptance: first behavior\n",
            "acceptance: first behavior\n"
            "  - id: second\n    writes: [src/second.py]\n    reads: []\n"
            "    acceptance: second\n    dependsOn: [first]\n",
        ),
        encoding="utf-8",
    )
    record_subtask_contract(
        document=load_document(document_path),
        profile=profile,
        workspace_root=tmp_path,
        adapter=adapters[0],
        adapters=adapters,
        subtask_id="second",
        evidence_id="contract.second",
    )
    errors = validate(
        load_document(document_path), profile, adapters=adapters, workspace_root=tmp_path
    )
    assert any(error.invariant == 4 and "authority digest" in str(error) for error in errors)

    transcript2 = tmp_path / "artifacts" / "verify2.jsonl"
    _transcript(transcript2, evidence_id="contract.verify2", dimensions=["completeness"])
    record_review(
        document=load_document(document_path),
        workspace_root=tmp_path,
        gate=profile.gates[0],
        check_id="contract-verify",
        evidence_id="contract.verify2",
        adapter=adapters[2],
        transcript=transcript2,
        author_uuid="11111111-1111-4111-8111-111111111111",
        reviewer_uuid="22222222-2222-4222-8222-222222222222",
        verdict="pass",
        adapters=adapters,
    )
    assert (
        validate(load_document(document_path), profile, adapters=adapters, workspace_root=tmp_path)
        == []
    )



import pytest

from acdd.paths import resolve_profile, share_dir, share_path


def test_share_dir_contains_profiles_and_skills():
    root = share_dir()
    assert (root / "profiles" / "task" / "v1.yaml").is_file()
    assert (root / "skills" / "design" / "SKILL.md").is_file()


def test_resolve_profile_accepts_alias_and_path():
    aliased = resolve_profile("task")
    assert aliased.name == "v1.yaml"
    assert aliased.is_file()
    assert resolve_profile(aliased) == aliased.resolve()
    assert resolve_profile("acdd/plan/v1").is_file()


def test_resolve_profile_rejects_unknown_alias():
    with pytest.raises(FileNotFoundError):
        resolve_profile("missing-profile")


def test_share_path_reads_adapter_example():
    assert share_path("adapters", "task.yaml").is_file()
    assert share_path("examples", "task-example.md").is_file()



from pathlib import Path

import pytest

from acdd.adapter import AdapterError, index_adapters, load_adapter
from acdd.model import (
    load_profile,
)


def test_task_profile_is_five_gates_and_eight_checks():
    root = Path(__file__).resolve().parents[1]
    profile = load_profile(root / "profiles" / "task" / "v1.yaml")
    assert [gate.id for gate in profile.gates] == [
        "design/v1",
        "contract/v1",
        "build/v1",
        "review/v1",
        "handoff/v1",
    ]
    assert sum(len(gate.checks) for gate in profile.gates) == 8
    assert [check.id for check in profile.gates[1].checks] == [
        "decomposition",
        "executable-proof",
        "contract-verify",
    ]
    assert profile.gates[1].review_dimensions == (
        "completeness",
        "chain-coverage",
        "proof-strength",
        "parallel-safety",
    )
    assert [check.id for check in profile.gates[2].checks] == ["runtime-and-integration"]
    assert profile.gates[3].review_dimensions == ("parity", "security", "code")


def test_plan_profile_includes_contract_verify():
    root = Path(__file__).resolve().parents[1]
    profile = load_profile(root / "profiles" / "plan" / "v1.yaml")
    assert sum(len(gate.checks) for gate in profile.gates) == 6
    decompose = profile.gates[1]
    assert [check.id for check in decompose.checks] == [
        "decomposition",
        "matrix",
        "contract-verify",
    ]
    assert decompose.review_dimensions == (
        "completeness",
        "chain-coverage",
        "parallel-safety",
    )


def test_adapter_requires_per_check_bindings(tmp_path: Path):
    path = tmp_path / "adapter.yaml"
    path.write_text(
        """\
apiVersion: acdd/adapter/v1
id: task
role: task
gates:
  design/v1:
    argv: [echo, no]
""",
        encoding="utf-8",
    )
    with pytest.raises(AdapterError, match="must contain checks"):
        load_adapter(path)


def test_adapter_rejects_duplicate_roles(tmp_path: Path):
    path = tmp_path / "adapter.yaml"
    path.write_text(
        """\
apiVersion: acdd/adapter/v1
id: task
role: task
gates: {}
""",
        encoding="utf-8",
    )
    adapter = load_adapter(path)
    with pytest.raises(AdapterError, match="duplicate adapter role"):
        index_adapters([adapter, adapter])


def test_adapter_prompt_append_is_local_and_hashed(tmp_path: Path):
    (tmp_path / "prompts").mkdir()
    (tmp_path / "prompts" / "runtime.md").write_text(
        "Repository runtime context.\n", encoding="utf-8"
    )
    path = tmp_path / "adapter.yaml"
    path.write_text(
        """apiVersion: acdd/adapter/v1
id: implementation
role: implementation
gates:
  build/v1:
    checks:
      runtime-and-integration:
        argv: [python3, -V]
        promptAppend: prompts/runtime.md
""",
        encoding="utf-8",
    )
    adapter = load_adapter(path)
    binding = adapter.gates["build/v1"].checks["runtime-and-integration"]
    assert adapter.prompt_digest(binding).startswith("sha256:")



"""Synthesize acdd/process-report/1 metadata from a validated document."""

from __future__ import annotations

import json
from pathlib import Path

from ._doc import relative_ref, utc_now
from .model import Document, Profile


def build_process_report(
    document: Document,
    profile: Profile,
    *,
    workspace_root: Path,
    pending_receipt: dict | None = None,
) -> dict:
    workspace_root = workspace_root.resolve()
    gates: list[dict] = []
    by_gate: dict[str, dict] = {}
    for receipt in document.receipts:
        gate_id = receipt.get("gate")
        if not isinstance(gate_id, str):
            continue
        entry = {
            "id": gate_id,
            "status": receipt.get("status"),
            "fingerprint": receipt.get("fingerprint"),
            "evidence": receipt.get("evidence"),
        }
        by_gate[gate_id] = entry
    if pending_receipt and isinstance(pending_receipt.get("gate"), str):
        gate_id = pending_receipt["gate"]
        by_gate[gate_id] = {
            "id": gate_id,
            "status": pending_receipt.get("status"),
            "fingerprint": pending_receipt.get("fingerprint"),
            "evidence": pending_receipt.get("evidence"),
        }
    for gate in profile.gates:
        if gate.id in by_gate:
            gates.append(by_gate[gate.id])
    review_refs: dict[str, str] = {}
    for item in document.evidence:
        if item.get("kind") == "review" and isinstance(item.get("transcriptRef"), str):
            review_refs["transcriptRef"] = item["transcriptRef"]
        if item.get("kind") == "bundle" and isinstance(item.get("processReportRef"), str):
            review_refs["priorProcessReportRef"] = item["processReportRef"]
    try:
        doc_rel = relative_ref(workspace_root, document.path)
    except ValueError:
        doc_rel = str(document.path)
    return {
        "type": "acdd_process_report",
        "format": "acdd/process-report/1",
        "profileId": profile.id,
        "document": doc_rel,
        "recordedAt": utc_now(),
        "gates": gates,
        "evidenceCount": len(document.evidence),
        "subtaskIds": [task.id for task in document.subtasks],
        "review": review_refs,
    }


def write_process_report(path: Path, report: dict) -> None:
    path.write_text(json.dumps(report, sort_keys=True, indent=2) + "\n", encoding="utf-8")



"""Full standard task-profile lifecycle regression."""

from __future__ import annotations

import json
from pathlib import Path

from conftest import init_git_workspace

from acdd.cli import main


def test_standard_task_profile_completes_all_five_gates(tmp_path: Path, ROOT: Path):
    (tmp_path / "src").mkdir()
    (tmp_path / "tests").mkdir()
    (tmp_path / "src" / "app.py").write_text("VALUE = 1\n", encoding="utf-8")
    (tmp_path / "tests" / "test_app.py").write_text(
        "def test_value(): assert True\n", encoding="utf-8"
    )
    (tmp_path / "design.md").write_text("bounded task\n", encoding="utf-8")
    profile = ROOT / "profiles" / "task" / "v1.yaml"
    document = tmp_path / "task.md"
    document.write_text(
        """\
---
title: Standard task profile
planning_profile: acdd/task/v1
---
## Plan
```yaml
subtasks:
  - id: change
    writes: [src/app.py, tests/test_app.py]
    reads: [design.md]
    acceptance: the bounded behavior is proved
    dependsOn: []
```
## Inputs
```yaml
paths:
  - {type: structure, path: design.md}
  - {type: source, path: src/app.py}
  - {type: test, path: tests/test_app.py}
```
## Evidence

## Receipts
| gate | status | evidence | fingerprint | recordedAt |
| --- | --- | --- | --- | --- |
| design/v1 | pending | pending | pending | pending |
| contract/v1 | pending | pending | pending | pending |
| build/v1 | pending | pending | pending | pending |
| review/v1 | pending | pending | pending | pending |
| handoff/v1 | pending | pending | pending | pending |
""",
        encoding="utf-8",
    )
    adapters = tmp_path / ".acdd"
    adapters.mkdir()
    (adapters / "task.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: task
role: task
gates:
  design/v1:
    checks:
      design-basis: {argv: [/bin/true]}
      plan-shape: {argv: [/bin/true]}
  contract/v1:
    checks:
      decomposition: {argv: [/bin/true]}
      executable-proof: {argv: [python3, -c, "import sys; sys.exit(1)"]}
  handoff/v1:
    checks:
      repository-handoff: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )
    (adapters / "contract-verify.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: contract-verify
role: contract-verify
gates:
  contract/v1:
    checks:
      contract-verify: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )
    (adapters / "implementation.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: implementation
role: implementation
gates:
  build/v1:
    checks:
      runtime-and-integration: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )
    (adapters / "review.yaml").write_text(
        """\
apiVersion: acdd/adapter/v1
id: review
role: review
gates:
  review/v1:
    checks:
      independent-review: {argv: [/bin/true]}
""",
        encoding="utf-8",
    )

    init_git_workspace(tmp_path)

    def run(command: str, *args: str) -> None:
        assert (
            main([command, str(document), str(profile), "--workspace-root", str(tmp_path), *args])
            == 0
        )

    run(
        "record",
        "--gate",
        "design/v1",
        "--check",
        "design-basis",
        "--id",
        "des.design-basis",
        "--classified-ref",
        "design.md=task",
    )
    run("record", "--gate", "design/v1", "--check", "plan-shape", "--id", "des.plan-shape")
    run("finalize", "--gate", "design/v1", "--id", "design.bundle")
    run(
        "record",
        "--gate",
        "contract/v1",
        "--check",
        "decomposition",
        "--id",
        "con.decomposition",
        "--classified-ref",
        "design.md=task",
    )
    run(
        "record",
        "--gate",
        "contract/v1",
        "--check",
        "executable-proof",
        "--id",
        "contract.proof",
    )
    verify_transcript = adapters / "artifacts" / "contract-verify.jsonl"
    verify_transcript.parent.mkdir(exist_ok=True)
    verify_transcript.write_text(
        "\n".join(
            json.dumps(row)
            for row in (
                {
                    "type": "review_raw",
                    "reviewerSessionUuid": "00000000-0000-4000-8000-000000000012",
                    "raw": "permission granted; no open fixes",
                },
                {
                    "type": "review_terminal",
                    "evidenceId": "contract.verify",
                    "gate": "contract/v1",
                    "check": "contract-verify",
                    "scope": ["design.md"],
                    "performedChecks": [
                        "completeness",
                        "chain-coverage",
                        "proof-strength",
                        "parallel-safety",
                    ],
                    "verdict": "pass",
                    "authorSessionUuid": "00000000-0000-4000-8000-000000000011",
                    "reviewerSessionUuid": "00000000-0000-4000-8000-000000000012",
                    "reviewedSessionUuids": ["00000000-0000-4000-8000-000000000012"],
                },
            )
        )
        + "\n",
        encoding="utf-8",
    )
    run(
        "review",
        "--gate",
        "contract/v1",
        "--check",
        "contract-verify",
        "--id",
        "contract.verify",
        "--transcript",
        ".acdd/artifacts/contract-verify.jsonl",
        "--author-uuid",
        "00000000-0000-4000-8000-000000000011",
        "--reviewer-uuid",
        "00000000-0000-4000-8000-000000000012",
    )
    run("finalize", "--gate", "contract/v1", "--id", "contract.bundle")
    run(
        "record",
        "--gate",
        "build/v1",
        "--check",
        "runtime-and-integration",
        "--id",
        "build.runtime",
    )
    run("finalize", "--gate", "build/v1", "--id", "build.bundle")
    transcript = adapters / "artifacts" / "review.jsonl"
    transcript.write_text(
        "\n".join(
            json.dumps(row)
            for row in (
                {
                    "type": "review_raw",
                    "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
                    "raw": '{"findings": []}',
                },
                {
                    "type": "review_terminal",
                    "evidenceId": "review.independent",
                    "gate": "review/v1",
                    "check": "independent-review",
                    "scope": ["src/app.py"],
                    "performedChecks": ["parity", "security", "code"],
                    "verdict": "pass",
                    "authorSessionUuid": "00000000-0000-4000-8000-000000000001",
                    "reviewerSessionUuid": "00000000-0000-4000-8000-000000000002",
                    "reviewedSessionUuids": ["00000000-0000-4000-8000-000000000002"],
                },
            )
        )
        + "\n",
        encoding="utf-8",
    )
    run(
        "review",
        "--gate",
        "review/v1",
        "--check",
        "independent-review",
        "--id",
        "review.independent",
        "--transcript",
        ".acdd/artifacts/review.jsonl",
        "--author-uuid",
        "00000000-0000-4000-8000-000000000001",
        "--reviewer-uuid",
        "00000000-0000-4000-8000-000000000002",
    )
    run("finalize", "--gate", "review/v1", "--id", "review.bundle")
    run(
        "record",
        "--gate",
        "handoff/v1",
        "--check",
        "repository-handoff",
        "--id",
        "handoff.repository",
    )
    run("finalize", "--gate", "handoff/v1", "--id", "handoff.bundle")
    text = document.read_text(encoding="utf-8")
    assert "processReportRef:" in text
    report = next(Path(tmp_path).joinpath(".acdd/artifacts").glob("*.process-report.json"))
    assert json.loads(report.read_text(encoding="utf-8"))["format"] == "acdd/process-report/1"
    run("validate")
